#updated 20221221
#oagOutReportTxt.ps1

function Get-TextReportPath([string]$FileName) {
#updated 20211221
	[string]$sExportPath = ''
	[string]$sOldFullName = ''

	$sExportPath = Join-Path -Path $CurrentExtensionDataPath -ChildPath 'Reports'
	# mkdir
	$null = New-Item -ItemType Directory -Path $sExportPath -Force

	if ($FileName) {
		$sExportPath = Join-Path -Path $sExportPath -ChildPath "$FileName.txt"

		# relocate file created by older code
		$sOldFullName = $sExportPath.replace('\Reports\','\')
		if ([System.IO.File]::Exists($sOldFullName)) {
			$null = [System.IO.File]::Move($sOldFullName, $sExportPath)
			# if move fails, delete
			if ([System.IO.File]::Exists($sOldFullName)) {
				$null = Remove-Item -LiteralPath $sOldFullName
			}
		}
	}

	return $sExportPath
}

function Report-AllGamesContainingMatchesOnTag1_AND_Tag2(
	[string]$sToMatch1,
	[string]$sToMatch2,
	[string]$sDescAndTxtFileName
) {

	[string]$sMsg = ''
	[string]$fullnameReport = ''
	[int]$iCount = 0
	[int]$iMaxRowsDisplay = 45
	$sbListNewLined = [System.Text.StringBuilder]''

	# trap invalid param
	if ((!$sToMatch1) -or (!$sToMatch2) -or (!$sDescAndTxtFileName)) {
		$__logger.Error("missing params for function 'Report-AllGamesContainingMatchesOnTag1_AND_Tag2'")
		exit
	}

	$matchedGames = $PlayniteApi.Database.Games | Where-Object { ($_.Tags -match $sToMatch1) -and ($_.Tags -match $sToMatch2) } | Sort-Object $_.name
	$iCount = $($matchedGames.count)

	ForEach ($game in $matchedGames) {
		$null = $sbListNewLined.AppendLine(" $($game.name) ($($game.Platform.name))")
	}

	$fullnameReport = Get-TextReportPath -FileName $sDescAndTxtFileName

	# output report
	if ($iCount -gt 0) {
		("`r`n$iCount $sDescAndTxtFileName `r`n`r`n" + $($sbListNewLined.ToString()) + "`r`n") | Out-File $fullnameReport -Encoding 'UTF8'
	}

	# display results
	$sMsg = "$iCount $sDescAndTxtFileName`r`n"
	if ($iCount -gt 0) {
		if ($iCount -gt $iMaxRowsDisplay) {
			$sMsg = $sMsg + "`r`nThere are too many to display. Check the output in the ExtensionData folder to see the list."
		} else {
			$sMsg = $sMsg + "`r`nSee list below: (This is also written to a file in the ExtensionData folder)`r`n`r`n" + $($sbListNewLined.ToString())
		}
	}
	$null = $PlayniteApi.Dialogs.ShowMessage($sMsg, 'RESULTS')

	# open/delete report
	if ([System.IO.File]::Exists($fullnameReport)) {
		if ($iCount -gt 0) {
			$null = Invoke-Item -LiteralPath $fullnameReport
		} else {
			$null = Remove-Item -LiteralPath $fullnameReport
		}
	}
}