#updated 20221006
#oagStringsGames.ps1

function Convert-ExpandTwoDigitYearsInGameStrings([string]$value) {
#description: fix game names ending with two digit year
#updated 20221006

	if ($value -match ' 9') {
		if ($value -match '9[89]$') {
			$value = $value.replace(' 9',' 199')
		}
	} elseif ($value -match "'9[0-9]") {
		$value = $value.replace("'9",'199')
	}

	return $value
}

function Convert-RomanNumeralsToArabicInGameStrings([string]$value) {
#notes: it is impossible to achieve 100% accuracy on roman numeral conversion from title strings
#updated 20220928 added sort improvement for FF & DQ
#updated 20220221 'Final Fantasy (V|VI) Advance'
#updated 20211104 added additional fix for strings ending in ' I'

	[bool]$bContainsRomanNumerals = $false
	[string]$sFranchise = ''

	# set sFranchise
	if ($value -match 'Final Fantasy') {
		$sFranchise = 'Final Fantasy'
	} elseif ($value -match 'Dragon Quest') {
		$sFranchise = 'Dragon Quest'
	}

	# check if value contains Roman numerals
	if ( ($value -cmatch ' II') -or ($value -cmatch ' IV') -or ($value -cmatch ' IX') -or ($value -cmatch ' XV') ) {
		# detect instances of roman numerals: 2,3,4,9,15,16,17,18
		$bContainsRomanNumerals = $true
	} elseif ($value -match 'Episode I') {
		# roman numeral 'I' rarely appears in titles
		# 'I' usually appears in titles as 'I & II' (caught in above test) or as 'Episode I'
		$bContainsRomanNumerals = $true
	} elseif ($value -match 'III') {
		# detect instances of roman numerals: 3,8,13
		$bContainsRomanNumerals = $true
	} elseif ( ($value -cmatch 'VII') -or ($value -cmatch ' XII') -or ($value -cmatch ' XIV') ) {
		# detect instances of roman numerals: 7,8,12,13,14,17,18
		$bContainsRomanNumerals = $true
	} elseif ($sFranchise) {
		$bContainsRomanNumerals = $true
		# this test is incorrect but it helps capture all titles in these franchises that contain roman numerals
		# it has very low risk for franchise titles that don't contain roman numerals
	} elseif ( ($value -cmatch '\ V$') -or ($value -cmatch ' V:') -or ($value -cmatch '\ VI$') -or ($value -cmatch ' VI:') ) {
		# this test has high risk of including false positives
		# detect instances of roman numerals: 5,6
		$bContainsRomanNumerals = $true
	} elseif ($value -cmatch '\ I$') {
		# roman numeral 'I' rarely appears in titles
		# this test has a very low risk of false positives
		$bContainsRomanNumerals = $true
	} else {
		# no testing for 10 and 11 due to high number of false positives
		# FF + DQ are already covered by earlier test
	}

	if ($bContainsRomanNumerals -eq $true) {
		if ($value -cmatch ' XV') {
			$value = $value.replace(' XVII',' 17')
			$value = $value.replace(' XVI',' 16')
			$value = $value.replace(' XV',' 15')
		} elseif ($value -cmatch ' XI') {
			$value = $value.replace(' XIV',' 14')
			$value = $value.replace(' XIII',' 13')
			$value = $value.replace(' XII',' 12')
			$value = $value.replace(' XI',' 11')
			} elseif ($value -cmatch ' X') {
			# 'X' rarely appears in titles to mean 10
			# this has highest risk of false positives

			# only convert FF and DQ for safety
			if ($sFranchise) {
				$value = $value.replace(' X',' 10')
			}
		} elseif ($value -cmatch ' IX') {
			$value = $value.replace(' IX',' 9')
		} elseif ($value -cmatch ' VI') {
			$value = $value.replace(' VIII',' 8')
			$value = $value.replace(' VII',' 7')
			$value = $value.replace(' VI',' 6')
		} else {

			if ($value -cmatch ' V') {
				$value = $value.replace(' V:',' 5:')
				if ($value -cmatch '\ V$') {
					$value = $value.replace(' V',' 5')
				}

				if ($value -match 'Final Fantasy') {
					$value = $value.replace(' VI ',' 6 ')
					$value = $value.replace(' V ',' 5 ')
				}
			}

			if ($value -cmatch ' IV') {
				$value = $value.replace(' IV',' 4')
			}

			if ($value -cmatch '\ I$') {
				$value = $value.Substring(0, $value.Length - 2) + ' 1'
			} elseif ($value -cmatch ' I') {
				$value = $value.replace('I & II','1 & 2')
				if ($value -cmatch ' II') {
					$value = $value.replace(' III',' 3')
					$value = $value.replace(' II',' 2')
				}
				# replace 2 and 3 before replacing episode (otherwise risk assign episode 2 as 1)
				$value = $value.replace('pisode I','pisode 1')
			}
		}
	}

	# improve sort order for large series by altering "series #" to "series 0#"
	if ($sFranchise) {
		if ($value -match "$sFranchise [2-9]") {
			$value = $value.replace("$sFranchise ","$sFranchise 0")
			$sFranchise = $sFranchise.ToUpper()
			$value = $value.replace("$sFranchise ","$sFranchise 0")
		}
	}

	return $value
}