function Get-ThisExtensionName {
	return 'Fix Game Names'
}

function OnApplicationStarted {
	if ([System.IO.File]::Exists("$CurrentExtensionDataPath\AUTORUN_ENABLED")) {
		$StopWatch = [system.diagnostics.stopwatch]::StartNew()
		$__logger.Info("auto run started")
		Invoke-RenameGamesToUndoTitleSort $true
		Invoke-AddSortingNameToGames $true
		$__logger.Info("auto run finished. running time $($StopWatch.Elapsed)")
	}
}

function GetGameMenuItems() {
	param($menuArgs)

	if (Get-IsDebugMode -eq $true) {

		$menuItem1 = New-Object Playnite.SDK.Plugins.ScriptGameMenuItem
		$menuItem1.Description = 'Display calculated Sorting Name for selected game'
		$menuItem1.FunctionName = 'Get-CalculatedSortingNameForSelectedGame'
		$menuItem1.MenuSection = 'DEV TESTS'

		return $menuItem1
	}
}

function GetMainMenuItems() {
	param($menuArgs)

	[string]$sMenuSection = 'Fix Game Names'

	$menuItem1 = New-Object Playnite.SDK.Plugins.ScriptMainMenuItem
	$menuItem1.Description = 'Name field - Trim'
	$menuItem1.FunctionName = 'Invoke-RenameGamesTrimNames 0'
	$menuItem1.MenuSection = "@$sMenuSection"

	$menuItem2 = New-Object Playnite.SDK.Plugins.ScriptMainMenuItem
	$menuItem2.Description = 'Name field - Remove Title Sort style'
	$menuItem2.FunctionName = 'Invoke-RenameGamesToUndoTitleSort 0'
	$menuItem2.MenuSection = "@$sMenuSection"

	$menuItem3 = New-Object Playnite.SDK.Plugins.ScriptMainMenuItem
	$menuItem3.Description = 'Sorting Name field - Populate'
	$menuItem3.FunctionName = 'Invoke-AddSortingNameToGames 0'
	$menuItem3.MenuSection = "@$sMenuSection"

	$menuItem4 = New-Object Playnite.SDK.Plugins.ScriptMainMenuItem
	$menuItem4.Description = 'Sorting Name field - Remove any matching game name'
	$menuItem4.FunctionName = 'Invoke-RemoveRedundantSortingNames'
	$menuItem4.MenuSection = "@$sMenuSection"

	$menuItem5 = New-Object Playnite.SDK.Plugins.ScriptMainMenuItem
	$menuItem5.Description = 'Sorting Name field - Trim'
	$menuItem5.FunctionName = 'Invoke-TrimSortingNames'
	$menuItem5.MenuSection = "@$sMenuSection"

	$menuItem6 = New-Object Playnite.SDK.Plugins.ScriptMainMenuItem
	$menuItem6.Description = 'Sorting Name field - Apply Title Sort to existing names'
	$menuItem6.FunctionName = 'Invoke-RenameSortingNamesToApplyTitleSort'
	$menuItem6.MenuSection = "@$sMenuSection"

	$menuItem7 = New-Object Playnite.SDK.Plugins.ScriptMainMenuItem
	$menuItem7.Description = 'Sorting Name field - Report existing names not matching calculated values'
	$menuItem7.FunctionName = 'Invoke-ReportSortingNamesNotMatchCalculated'
	$menuItem7.MenuSection = "@$sMenuSection"

	$menuItem8 = New-Object Playnite.SDK.Plugins.ScriptMainMenuItem
	$menuItem8.Description = 'Sorting Name field - Update existing names to calculated values (with prompt)'
	$menuItem8.FunctionName = 'Invoke-ReviewSortingNamesNotMatchCalculated'
	$menuItem8.MenuSection = "@$sMenuSection"

	$menuItem9 = New-Object Playnite.SDK.Plugins.ScriptMainMenuItem
	$menuItem9.Description = 'Sorting Name field - Update existing names to calculated values (Only automatic fixes. Does not fix any needing review.)'
	$menuItem9.FunctionName = 'Invoke-ReviewSortingNamesNotMatchCalculatedAUTO'
	$menuItem9.MenuSection = "@$sMenuSection"

	return $menuItem1, $menuItem2, $menuItem3, $menuItem4, $menuItem5, $menuItem6, $menuItem7, $menuItem8, $menuItem9
}

function Get-IsDebugMode {
	[bool]$bResult = ([System.IO.File]::Exists("$CurrentExtensionDataPath\DEBUG"))
	return $bResult
}

function Clean-SortingName([string]$value) {


	# load functions
	. "$CurrentExtensionInstallPath\lib\oagStringsGames.ps1"
	. "$CurrentExtensionInstallPath\lib\oagStringsGen.ps1"


	[bool]$bDebug = (Get-IsDebugMode -eq $true)

	$value = $value.TrimEnd().TrimStart()
	$value = Convert-RomanNumeralsToArabicInGameStrings -value $value
	$value = Convert-ExpandTwoDigitYearsInGameStrings -value $value

	# fix Driv3r
	if ($value -match '3') {
		$value = $value.replace('Driv3r','Driver 3')
	}

	$value = Convert-ApplyTitleSort -value $value -bEngOnly $false

	$value = $value.replace('™','').replace('®','')

# @tbw question next line
	$value = $value.replace('Ⓡ','')

	$value = $value.TrimEnd()

	return $value
}

function Invoke-RenameGamesTrimNames([bool]$bSilent = $false) {
	Invoke-CallStackNames -bActionIsTrim $true -bSilent $bSilent
}

function Invoke-RenameGamesToUndoTitleSort([bool]$bSilent = $false) {
	Invoke-CallStackNames -bActionIsTrim $false -bSilent $bSilent
}

function Invoke-CallStackNames([bool]$bActionIsTrim, [bool]$bSilent = $false) {


	# load functions
	. "$CurrentExtensionInstallPath\lib\oagStringsGen.ps1"


	[int]$iCountNeedFixTotal = 0
	[int]$iCountFixedTotal = 0
	[string]$sNewValue = ''
	[string]$sMsg = ''
	$sbListNewLined = [System.Text.StringBuilder]''
	[int]$iMaxRowsDisplay = 45

	if ($bActionIsTrim -eq $true) {
		$matchedGames = $PlayniteApi.Database.Games | Where-Object { ($_.name -match '\ $') -or ($_.name -match '^ ') } | Sort-Object $_.name
	} else {

		# english language only
		#$matchedGames = $PlayniteApi.Database.Games | Where-Object { ($_.name -cmatch ', The') -or ($_.name -cmatch '\, A$') } | Sort-Object $_.name

		# include support for titles in languages other than english
		#$matchedGames = $PlayniteApi.Database.Games | Where-Object { ($_.name -cmatch ', The') -or ($_.name -cmatch '\, A$') -or ($_.name -cmatch '\, Der$') -or ($_.name -cmatch '\, Die$') -or ($_.name -cmatch '\, La$') -or ($_.name -cmatch '\, Les$') -or ($_.name -cmatch '\, El$') } | Sort-Object $_.name

		# detect match at any position in string
		$matchedGames = $PlayniteApi.Database.Games | Where-Object { ($_.name -cmatch ', The') -or ($_.name -cmatch ', A') -or ($_.name -cmatch ', Der') -or ($_.name -cmatch ', Die') -or ($_.name -cmatch ', La') -or ($_.name -cmatch ', Les') -or ($_.name -cmatch ', El') } | Sort-Object $_.name
	}

	$iCountNeedFixTotal = $matchedGames.count


	ForEach ($game in $matchedGames) {

		$sNewValue = ''

		if ($bActionIsTrim -eq $true) {
			$sNewValue = ($game.name).TrimEnd().TrimStart()
		} else {
			$sNewValue = Convert-ApplyTitleSort -value $($game.name) -bEngOnly $false -bInvert $true
		}

		if (($sNewValue) -and ($game.name -ne $sNewValue)) {
			if ($iCountFixedTotal -lt $iMaxRowsDisplay) {
				$null = $sbListNewLined.AppendLine("$game")
			}
			$__logger.Info("rename game from '$game' ($($game.Platform.name)) to '$sNewValue'")
			$iCountFixedTotal++

			$game.name = $sNewValue
			$game.Modified = (Get-Date)
			$null = $PlayniteApi.Database.Games.Update($game)
		} else {
			$__logger.Debug("FALSE POSITIVE '$game' ($($game.Platform.name)) '$sNewValue'")
			$iCountNeedFixTotal--
		}
	}


	# display results
	$sMsg = "$iCountNeedFixTotal games needed fixing."
	if ($iCountNeedFixTotal -gt 0) {
		$sMsg = $sMsg + "`r`n$iCountFixedTotal games were fixed."
		if ($iCountFixedTotal -lt $iMaxRowsDisplay) {
			$sMsg = $sMsg + "`r`n`r`n" + $($sbListNewLined.ToString())
		}
	}

	if ($bSilent -eq $false) {
		$null = $PlayniteApi.Dialogs.ShowMessage($sMsg, $(Get-ThisExtensionName))
	}
}

function Get-CalculatedSortingNameForSelectedGame {


	# load functions
	. "$CurrentExtensionInstallPath\lib\oagGetSelectedGame.ps1"


	[string]$sNewValue = ''
	[string]$sMsg = ''

	# only allow one game
	$game = $(Get-OneSelectedGame)

	$sNewValue = $game.name
	$sNewValue = Clean-SortingName -value $sNewValue

	$sMsg = $sNewValue

	if ($game.SortingName) {
		if ($($game.SortingName) -eq $sNewValue) {
			$sMsg = $sMsg + "`r`n`r`n`r`n" + "The calculated value matches the existing Sorting Name"
		} else {
			$sMsg = $sMsg + "`r`n`r`n`r`n" + "The calculated value does not match the existing Sorting Name:" + "`r`n" +"    $($game.SortingName)"
		}
	}

	$null = $PlayniteApi.Dialogs.ShowMessage($sMsg, $(Get-ThisExtensionName))
}

function Invoke-AddSortingNameToGames([bool]$bSilent = $false) {
	Invoke-CallStackSortingNames -iSwitch 1 -bSilent $bSilent
}

function Invoke-RemoveRedundantSortingNames {
	Invoke-CallStackSortingNames -iSwitch 2
}

function Invoke-TrimSortingNames {
	Invoke-CallStackSortingNames -iSwitch 3
}

function Invoke-RenameSortingNamesToApplyTitleSort {
	Invoke-CallStackSortingNames -iSwitch 4
}

function Invoke-ReviewSortingNamesNotMatchCalculated {
	Invoke-CallStackSortingNames -iSwitch 5
}

function Invoke-ReviewSortingNamesNotMatchCalculatedAUTO {
	Invoke-CallStackSortingNames -iSwitch 7
}

function Invoke-ReportSortingNamesNotMatchCalculated {
	Invoke-CallStackSortingNames -iSwitch 6
}

function Invoke-CallStackSortingNames([parameter(mandatory)][int]$iSwitch, [bool]$bSilent = $false) {


	# load functions
	. "$CurrentExtensionInstallPath\lib\oagStringsGames.ps1"
	. "$CurrentExtensionInstallPath\lib\oagStringsGen.ps1"
	. "$CurrentExtensionInstallPath\lib\oagOutReportTxt.ps1"


	[int]$iCountNeedFixTotal = 0
	[int]$iCountFixedTotal = 0
	[string]$sNewValue = ''
	[string]$sMsg = ''
	[int]$iMaxRowsDisplay = 45
	[int]$iCountAutoApproved = 0
	[bool]$bApplyAction = $false
	[bool]$bSkipRemaining = $false
	[bool]$bReportOnly = $false
	[bool]$bAskUser = $true
	[bool]$bDontPromptButOnlyFixSafe = $false
	[string]$sExportPath = ''

	[System.Collections.Generic.List[Object]]$colGamesProcessed = @()

	switch ($iSwitch) {
		6 {
			$bReportOnly = $true
			$iSwitch = 5
		}
		7 {
			$bDontPromptButOnlyFixSafe = $true
			$iSwitch = 5		
		}
	}

	$matchedGames = switch ($iSwitch) {
		1 { $PlayniteApi.Database.Games | Where-Object { (!$_.sortingname) } | Sort-Object $_.name }
		2 { $PlayniteApi.Database.Games | Where-Object { ($_.sortingname -eq $_.name) } | Sort-Object $_.name }
		3 { $PlayniteApi.Database.Games | Where-Object { ($_.sortingname) -and ($_.sortingname.EndsWith(' ')) } | Sort-Object $_.name }
		4 {
			# english language only
			#$PlayniteApi.Database.Games | Where-Object { ( $_.sortingname -match ('^The ') -or $_.sortingname -match ('^A '))

			# include support for languages other than english
			$PlayniteApi.Database.Games | Where-Object { ( $_.sortingname -match ('^The ') -or $_.sortingname -match ('^A ') -or $_.sortingname -match ('^Der ') -or $_.sortingname -match ('^Die ') -or $_.sortingname -match ('^La ') -or $_.sortingname -match ('^Les ') -or $_.sortingname -match ('^El ')
			)
		} | Sort-Object $_.name }
		5 { $PlayniteApi.Database.Games | Where-Object { ($_.sortingname) } | Sort-Object $_.name }
	}

	$sMsg = switch ($iSwitch) {
		1 { "sorting name updated for game '{0}' - set as '{1}'" }
		2 { "sorting name updated for game '{0}' - removed '{1}'" }
		3 { "sorting name updated for game '{0}' - trimmed '{1}'" }
		4 { "sorting name updated for game '{0}' - applied Title Sort to existing sorting name '{1}'" }
	}

	$iCountNeedFixTotal = switch ($iSwitch) {
		1 {0}
		5 {0}
		default {$matchedGames.count}
	}

	ForEach ($game in $matchedGames) {

		# defaults
		$bApplyAction = $false
		$bAskUser = $true

		# get $sNewValue
		switch ($iSwitch) {
			1 {
				$sNewValue = $game.name
				$sNewValue = Clean-SortingName -value $sNewValue
			}
			2 {$sNewValue = ''}
			3 {$sNewValue = $game.sortingname.TrimEnd()}
			4 {
				$sNewValue = $game.sortingname
				$sNewValue = $sNewValue.TrimEnd()
				$sNewValue = Convert-ApplyTitleSort -value $sNewValue -bEngOnly $false
				#$sNewValue = $sNewValue.replace('™','').replace('®','')
				$sNewValue = $sNewValue.TrimEnd()
			}
			5 {
				$sNewValue = $game.name
				$sNewValue = Clean-SortingName -value $sNewValue
				#$sNewValue = $sNewValue.replace('’',"'")
			}
		}

		# verify ok to proceed
		switch ($iSwitch) {
			1 {
				if (($sNewValue) -and ($sNewValue -ne $game.name)) {

					# prevent adding SortingName that matches the game name
					$bApplyAction = $true
					if ($sNewValue -eq $game.name) {
						$bApplyAction = $false
					} elseif ($game.name -match '\™$') {
						if ($sNewValue -eq $($($game.name).replace('™','').TrimEnd())) {
							$bApplyAction = $false
						}
					}

					if ($bApplyAction -eq $true) {
						$iCountNeedFixTotal++
					}
				}
			}
			5 {
				if (($sNewValue) -and ($sNewValue -ne $game.sortingname)) {

					$iCountNeedFixTotal++

					# prevent adding SortingName that matches the game name
					$bApplyAction = $true
					if ($sNewValue -eq $game.name) {
						$bApplyAction = $false
						} elseif ($($game.name).EndsWith('™')) {
						if ($sNewValue -eq $($($game.name).replace('™','').TrimEnd())) {
							$bApplyAction = $false
						}
					}
					if ($bApplyAction -eq $false) {
						if ($bReportOnly -eq $true) {
							$sNewValue = ''
						} else {
							$sNewValue = "'' (Playnite will treat this the same as the game name for sorting purposes)"
						}
					}

					if ($bReportOnly -eq $true) {

						# reset
						$hashGameData = $null
						$oCustom = $null

						$hashGameData = @{
							Name = ($game.name)
							Platform = ($game.Platform.name)
							SortingName = ($game.SortingName)
							CalculatedSortingName = $sNewValue
							DatabaseId = ($game.Id)
						}

						$oCustom = [PsCustomObject]$hashGameData
						$null = $colGamesProcessed.Add($oCustom)
					} else {
						# check if can automatic approve to minimise number of user approvals required
						$sMsg = ''
						if ($sNewValue.StartsWith($($game.sortingname))) {
							$sMsg = "automatically approve change from '$($game.sortingname)' to '$sNewValue' (starts with same string)"
						} elseif ("$(($sNewValue).replace(', The',''))" -eq "$(($game.sortingname).replace(', The',''))") {
							$sMsg = "automatically approve change from '$($game.sortingname)' to '$sNewValue' (only difference is article)"
						} elseif ("$(($sNewValue).replace(', A',''))" -eq "$(($game.sortingname).replace(', A',''))") {
							$sMsg = "automatically approve change from '$($game.sortingname)' to '$sNewValue' (only difference is article)"
						} elseif ("$(($sNewValue).replace(', Die',''))" -eq "$(($game.sortingname).replace(', Die',''))") {
							$sMsg = "automatically approve change from '$($game.sortingname)' to '$sNewValue' (only difference is article)"
						} elseif ($sNewValue -eq "$(($game.sortingname).replace('™','').replace('®',''))") {
							$sMsg = "automatically approve change from '$($game.sortingname)' to '$sNewValue' (only difference is TM and R symbols)"
						} elseif ($($game.sortingname) -match '5ol') {
							# scrap calculated value and redo to fix error from older version
							$sNewValue = $game.sortingname
							$sNewValue = $sNewValue.replace('5ol','Vol')
							$sNewValue = Convert-RomanNumeralsToArabicInGameStrings -value $sNewValue

							$sMsg = "automatically approve change from '$($game.sortingname)' to '$sNewValue' (correct error caused by old version of Convert-RomanNumeralsToArabicInGameStrings)"
						} elseif ($sNewValue -eq "$(($game.sortingname).replace(' - ',': '))") {
							$sMsg = "automatically approve change from '$($game.sortingname)' to '$sNewValue' (probably populated current SortingName before fixing game name from No-Intro)"
						} elseif ($sNewValue.replace(' 0',' ') -eq $($game.sortingname)) {
							$sMsg = "automatically approve change from '$($game.sortingname)' to '$sNewValue' (apply improvement from updated Convert-RomanNumeralsToArabicInGameStrings)"
						}
						if ($sMsg.length -gt 1) {
							# this is an auto approval
							$bAskUser = $false
							$bApplyAction = $true
							$__logger.Info($sMsg)
							$sMsg = ''
						} else {
							# will need manual approval
							$bApplyAction = $false

							if ($bDontPromptButOnlyFixSafe -eq $true) {
								$bAskUser = $false
								# only process fixes that can be automatically approved
							}
						}

						if ($bAskUser -eq $true) {
							# ask user
							$sMsg = "Game Name = '$($game.name)' ($($game.Platform.name))`r`nSorting Name = '$($game.sortingname)'`r`nCalculated Sorting Name = '$sNewValue'`r`n`r`nDo you want to update the Sorting Name to the new value?`r`n`r`n`r`n(Press 'Cancel' if you want to skip checking remaining games)"
							$UserResponse = $PlayniteApi.Dialogs.ShowMessage($sMsg, $(Get-ThisExtensionName), 3)
							switch ($UserResponse) {
								'Yes' {$bApplyAction = $true}
								'No' {
									$bApplyAction = $false
# POSSIBLE FUTURE IMPLEMENTATION write to config file to remember choice for future
								}
								'Cancel' {$bSkipRemaining = $true}
								default {$bSkipRemaining = $true}
							}
						}

						# user cancelled early
						if ($bSkipRemaining -eq $true) {
							$sMsg = "$iCountNeedFixTotal games were checked. All remaining games were skipped."
							$sMsg = $sMsg + "`r`n$iCountFixedTotal games were fixed."
							if ($iCountAutoApproved -gt 0) {
								$sMsg = $sMsg + "`r`n$iCountAutoApproved games were automatically approved for fixing."
							}
							if ($iCountFixedTotal -lt $iMaxRowsDisplay){
								$colGamesProcessed = $colGamesProcessed | Sort-Object $_.name

								if ($colGamesProcessed.name) {
									$sMsg = $sMsg + "`r`n`r`n" + [string]::Join("`r`n", $colGamesProcessed.name)
								}
							}
							$null = $PlayniteApi.Dialogs.ShowMessage($sMsg, $(Get-ThisExtensionName))
							exit
						}

						$sMsg = "sorting name updated for game '{0}' - replaced '{1}' with '{2}'"
						if ($sNewValue -eq "'' (Playnite will treat this the same as the game name for sorting purposes)") {
							$sNewValue = ''
						}
					}
				}
			}
			default {$bApplyAction = $true}
		}

		# update sortingname field
		if (($bApplyAction -eq $true) -and ($bReportOnly -ne $true)) {
			$null = $colGamesProcessed.Add($($game.name))

			switch ($iSwitch) {
				1 { $__logger.Info($("$sMsg" -f $game.name, $sNewValue)) }
				5 { $__logger.Info($("$sMsg" -f $game.name, $game.sortingname, $sNewValue)) }
				default { $__logger.Info($("$sMsg" -f $game.name, $game.sortingname)) }
			}

			$game.sortingname = $sNewValue
			$game.Modified = (Get-Date)
			$null = $PlayniteApi.Database.Games.Update($game)
			$iCountFixedTotal++
			if ($bAskUser -eq $false) {
				$iCountAutoApproved++
			}
		}
	}


	# display results
	$sMsg = "$iCountNeedFixTotal games needed fixing."
	if ($iCountNeedFixTotal -gt 0) {
		$sMsg = $sMsg + "`r`n$iCountFixedTotal games were fixed."
		if ($iCountAutoApproved -gt 0) {
			$sMsg = $sMsg + "`r`n$iCountAutoApproved games were automatically approved for fixing."
		}

		if ($($colGamesProcessed.count) -gt 0) {
			$colGamesProcessed = $colGamesProcessed | Sort-Object $_.name

			if ($iCountNeedFixTotal -lt $iMaxRowsDisplay){
				if ($colGamesProcessed.name) {
					$sMsg = $sMsg + "`r`n`r`n" + [string]::Join("`r`n", $colGamesProcessed.name)
				}
			}
		}
	}

	if ($bSilent -eq $false) {

		$null = $PlayniteApi.Dialogs.ShowMessage($sMsg, $(Get-ThisExtensionName))

		# report
		if ( ($iCountNeedFixTotal -gt 0) -and ($bReportOnly -eq $true)) {

			$sExportPath = Get-TextReportPath

			$sExportPath = Join-Path -Path $sExportPath -ChildPath 'Report games where Sorting Name does not match calculated Sorting Name.csv'

			$colGamesProcessed | Select Name, Platform, SortingName, CalculatedSortingName, DatabaseId | ConvertTo-Csv -NoTypeInformation | Out-File $sExportPath -Encoding 'UTF8'
			if ([System.IO.File]::Exists($sExportPath)) {
				$null = Invoke-Item -LiteralPath $sExportPath
			}
		}
	}
}